package net.knarcraft.bookswithoutborders.utility;

import java.util.ArrayList;
import java.util.List;

/**
 * A converter from an integer to a roman numeral
 */
public final class IntegerToRomanConverter {

    private final static List<Integer> romanValues = new ArrayList<>();
    private final static List<Character> romanCharacters = new ArrayList<>();

    static {
        // Initialize the roman numbers
        romanValues.add(1000);
        romanValues.add(500);
        romanValues.add(100);
        romanValues.add(50);
        romanValues.add(10);
        romanValues.add(5);
        romanValues.add(1);

        romanCharacters.add('M');
        romanCharacters.add('D');
        romanCharacters.add('C');
        romanCharacters.add('L');
        romanCharacters.add('X');
        romanCharacters.add('V');
        romanCharacters.add('I');
    }

    private IntegerToRomanConverter() {

    }

    /**
     * Gets the given number as a roman number string
     *
     * @param number <p>The number to convert</p>
     * @return <p>The roman representation of the number</p>
     */
    public static String getRomanNumber(int number) {
        StringBuilder output = new StringBuilder();
        int remainder = number;
        for (int i = 0; i < romanCharacters.size(); i++) {
            int romanValue = romanValues.get(i);
            char romanCharacter = romanCharacters.get(i);

            // Repeat the roman character, and calculate the new remainder
            if (remainder >= romanValue) {
                output.append(repeat(romanCharacter, remainder / romanValue));
                remainder = remainder % romanValue;
            }

            // Exit early to prevent unexpected trailing characters
            if (remainder == 0) {
                return output.toString();
            }

            // Generate the special case IV and similar
            for (int j = i; j < romanCharacters.size(); j++) {
                int value = romanValues.get(j);
                int difference = Math.max(romanValue - value, 0);
                
                /* If the remainder is "one" less than the current roman value, we hit the IV/IX/XL case.
                   Note that 5 triggers the special case when 10 is tested, as 5 = 10 - 5, which requires a test, so it
                   can be filtered out. */
                if (remainder == difference && value != romanValue / 2) {
                    output.append(romanCharacters.get(j)).append(romanCharacter);
                    remainder = 0;
                }
            }
        }

        return output.toString();
    }

    /**
     * Repeats the given character
     *
     * @param character <p>The character to repeat</p>
     * @param times     <p>The number of times to repeat the character</p>
     * @return <p>The repeated string</p>
     */
    private static String repeat(char character, int times) {
        return String.valueOf(character).repeat(Math.max(0, times));
    }

}
